/*
 *RobotControl.java
 *This holds the calculation code and the Main Operating Loop for the robot.
 *It receives commands from the window created by RobotInterface.java 
 */


/*
import gnu.io.CommPortIdentifier;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;*/

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.math.linear.LUDecompositionImpl;
import org.apache.commons.math.linear.RealMatrix;
import org.apache.commons.math.linear.RealMatrixImpl;



public class robotControl {
	//******************************************************************
	//*	CONTROL TABLE ADDRESSES (same names as in the driver from Servo magazine)
	//******************************************************************
	public static final int MODEL_NUMBER =  0x00;
	public static final int MODEL_NUMBER_H =  0x01;
	public static final int VERSION =  0x02;
	public static final int ID =  0x03;
	public static final int BAUD_RATE =  0x04;
	public static final int RETURN_DELAY_TIME =  0x05;
	public static final int CW_ANGLE_LIMIT =  0x06;
	public static final int CW_ANGLE_LIMIT_H =  0x07;
	public static final int CCW_ANGLE_LIMIT =  0x08;
	public static final int CCW_ANGLE_LIMIT_H =  0x09;
	public static final int RESERVED1 =  0x0A;
	public static final int LIMIT_TEMPERATURE =  0x0B;
	public static final int DOWN_LIMIT_VOLTAGE =  0x0C;
	public static final int UP_LIMIT_VOLTAGE =  0x0D;
	public static final int MAX_TORQUE =  0x0E;
	public static final int MAX_TORQUE_H =  0x0F;
	public static final int STATUS_RETURN_LEVEL =  0x10;	
	public static final int ALARM_LED =  0x11;
	public static final int ALARM_SHUTDOWN =  0x12;
	public static final int RESERVED2 =  0x13;
	public static final int DOWN_CALIBRATION =  0x14;
	public static final int DOWN_CALIBRATION_H =  0x15;
	public static final int UP_CALIBRATION =  0x16;
	public static final int UP_CALIBRATION_H =  0x17;
	public static final int TORQUE_ENABLE =  0x18;
	public static final int LED =  0x19;
	public static final int CW_COMPLIANCE_MARGIN =  0x1A;
	public static final int CCW_COMPLIANCE_MARGIN =  0x1B;
	public static final int CW_COMPLIANCE_SLOPE =  0x1C;
	public static final int CCW_COMPLIANCE_SLOPE =  0x1D;
	public static final int GOAL_POSITION =  0x1E; 
	public static final int GOAL_POSITION_H =  0x1F; 
	public static final int MOVING_SPEED =  0x20; 
	public static final int MOVING_SPEED_H =  0x21; 
	public static final int TORQUE_LIMIT =  0x22;
	public static final int TORQUE_LIMIT_H =  0x23; 
	public static final int PRESENT_POSITION =  0x24; 
	public static final int PRESENT_POSITION_H =  0x25; 
	public static final int PRESENT_SPEED =  0x26; 
	public static final int PRESENT_SPEED_H =  0x27; 
	public static final int PRESENT_LOAD =  0x28; 
	public static final int PRESENT_LOAD_H =  0x29; 
	public static final int PRESENT_VOLTAGE =  0x2A ;
	public static final int PRESENT_TEMPERATURE =  0x2B; 
	public static final int REGISTERED_INSTRUCTION =  0x2C; 
	public static final int RESERVE3 =  0x2D; 
	public static final int MOVING =  0x2E; 
	public static final int LOCK =  0x2F; 
	public static final int PUNCH =  0x30;
	public static final int PUNCH_H = 0x31;
	//********end control table addresses***********
	
    RobotInterface window;
    Timer MOLTimer;
    
    public static final int NUMLEGS = 6;
    public int[][] liveServoGoalPositions = new int[NUMLEGS][3];
    public boolean[][] servoGoalPositionChanged = new boolean[NUMLEGS][3];
    public int[][] servopluslimits = new int[NUMLEGS][3];
    public int[][] servominuslimits = new int[NUMLEGS][3];
    public double[][] measuredJointAngles = new double[NUMLEGS][3];//joint angles in radians
    public double[][] measuredJointTorques = new double[NUMLEGS][3];//joint torques in N.m
    public double[][] measuredForces = new double[NUMLEGS][3]; //forces on feet, derived from joint torque
    public double[][] measuredForcesFiltered = new double[NUMLEGS][3]; //forces on feet, derived from joint torque
    
    //Assumed full torque. AX-12 datasheet says it outputs 12 Kg.cm with 7V power, or 16.5 Kg.cm at 10V 
    //public static final double FULLTORQUENM = 12*9.806/100;  // <-- at 7V
    public static final double FULLTORQUENM = 16.5*9.806/100;  // <-- at 10V
    
    public static final double TORQUEACCURACY = .075;//experimentally, AX-12's reported fractional torque will vary by this amount. Changes within this
    												//are not to be trusted.
    public double[][] measuredForceError = new double[NUMLEGS][3]; /**change in force caused by perturbing torque by the above accuracy. 
    											* This changes with joint position and can be large , so you MUST account for it when making decisions.
    											* For example, forces within "measuredForceError" of 0 are often not real.    												
    											*/
    public boolean[][] measuredForceIsCertainlyNonzero = new boolean[NUMLEGS][3]; //For convenience. This is set to "false" if the measured force is  
    												// within "measuredForceError" of 0 and therefore probably not real
    public boolean jacobianIsNonSingular = true;
    
    //the below arrays will be used to calculate pulse length given angle, in the form y=mx+b
    //a.k.a. pulse=gain*angle + zero
    public double[][] servozero = new double[NUMLEGS][3];//pulse length that produces angle=0
    public double[][] servogain = new double[NUMLEGS][3];//change in pulse length per change in degree
    //other control and calibration variables
    int calPulse1 = 0; //length of pulse that yielded calAngle1, according to the user.
    int calAngle1 = -999;
    int calPulse2 = 0;
    int calAngle2 = -999;
    public static final int mainloopperiod = 25;//time delay between runs of the main loop, in ms.
    //code used for the test motion sequence
    double elapsed;//seconds passed since last iteration through main operating loop
    boolean movementSequenceOn = false;
    int sequenceStep; //where the leg is in the movement sequence
    double sequenceTimer;//timer to next step in sequence

    /*//serial stuff:
    static Enumeration	      portList;
    static CommPortIdentifier portId;
    static SerialPort	      serialPort;
    static OutputStream       outputStream;
    static boolean	      outputBufferEmptyFlag = false;
*/

    
	
	
	
	
	
    //mainloop. This is run repeatedly by the timer set up in main().
    //It handles overall control of the robot.
    public void mainloop(){
    	elapsed = (double)mainloopperiod/1000;//assumes we're executing at our requested rate. Replace this with a timer if that becomes a bad assumption. 
        
        //-----------handle calibration if enabled-------------
        if(window.calOn() == true){
        	livecalibrate();
        }
        else{//if cal is off, clear any cal-related commands or else they'll be acted on all at once later.
            window.saveCal = false;
            window.setPlusLim = false;
            window.setMinusLim = false;
            window.calAngle = -999;
        }

        //--------------handle leg position testing command-------
        if(window.testMove == true){
            window.testMove = false;//reset button variable
            movementSequenceOn = false;//cancel any sequences in progress
            moveLegTo(window.getLeg(), window.getTestPosition());
            
        }

        //--------------handle single step command------------
        if(window.takeStep == true){
            window.takeStep = false;//reset button variable
            movementSequenceOn = true; //start the step sequence
            sequenceStep = 0;
            sequenceTimer = 0;
        }

        if(movementSequenceOn){
        	boolean finished = continueSequence();
        	if(finished)
        		movementSequenceOn = false;
        }
        window.setText2("zero: " + servozero[window.getLeg()][window.getServo()] + " gain: "
                + servogain[window.getLeg()][window.getServo()]);
        window.setText3(" -limit: " + servominuslimits[window.getLeg()][window.getServo()]
                + "+limit: " + servopluslimits[window.getLeg()][window.getServo()]);

        transmitDigitalCommand();
        
        readServoData();
        if(window.outputTestData == true){
        	window.outputTestData = false;
        	findForce(true);
        }else{
        	findForce(false);
        }
    }
	
	
	
	
	
	
	
	
    
    //----------Constructor---------------
    //the constructor takes my user input window as an argument
    //It starts running the mainloop() method every 50ms until the application closes.
    //mainloop() polls the window for inputs and responds to them.
    robotControl(RobotInterface inputwindow) {
        window = inputwindow;

        //openSerial();//uncomment this to set up the RxTx serial port. Not used since I have the USB2Dynamixel.
        
        //set up USB2Dynamixel:
        int res = dynamixel_jni.dxl_initialize();
        dynamixel_jni.dxl_set_baud(1);//1 means 1 Mbaud; I think this uses the same scheme as the AX-12 does.
		if (res == 1){
			System.out.println("Successfully opened USB2Dynamixel");
		}else{
			System.out.println("Failed to open USB2Dynamixel");
		}
		
		
		//Turn on LED briefly to demonstrate communication
		dynamixel_jni.dxl_write_byte(1, 25, 1);//on servo 1, turn on LED.
		try{
			Thread.sleep(100);//delay to make LED flash visible
		}catch(Exception e){}
		dynamixel_jni.dxl_write_byte(1, 25, 0);//turn off LED
		
		configServosForTests();//set to low torque mode, to avoid damage during calibration.
		
		

        
        //Initialize stuff
        initRobotData();//read in the robot joint lengths
        boolean calibrated = loadCalibration();//read in servo calibration and angle limits from the auto-generated file
        if(calibrated){
        	for(int i=0; i<liveServoGoalPositions.length;i++){
                for(int j=0; j<liveServoGoalPositions[i].length; j++){
                    liveServoGoalPositions[i][j] = (servominuslimits[i][j]+servopluslimits[i][j])/2;//set initial pulse length to middle of calibrated band.
                }
            }
        }else{//not calibrated yet, so we'll have to make assumptions.
	        for(int i=0; i<liveServoGoalPositions.length;i++){
	            for(int j=0; j<liveServoGoalPositions[0].length; j++){
	                liveServoGoalPositions[i][j] = 512;//set default pulse length to middle of possible band, as this is unlikely to hurt a servo.
	            }
	        }
        }
        
		//Set the MOL to run repeatedly. This should happen after all the initialization stuff.
		MOLTimer = new Timer();
		MOLTimer.schedule(new TimerTask() {
		    public void run() {   mainloop();   }
			}, 100,//delay before first run; doesn't really matter
		        mainloopperiod);//repeat period in ms
    }
    
    

    //This gets called when the window is closed, by a listener method that I set up.
    void close(){
    	MOLTimer.cancel();
    	try{
    		//wait for last MOL to execute before sending torque-off command, because only one thread at a time should be using USB2Dynamixel.
    		//Not doing this causes the program to freeze if two threads try to use the USB2Dynamixel at once.
    		Thread.sleep(mainloopperiod*2);
    	}catch(Exception e){};
    	
    	torqueoff();
    	
    	try{
    		Thread.sleep(25);//wait for torqueoff command to go through before terminating.
    	}catch(Exception e){};
    	
    	dynamixel_jni.dxl_terminate();
    }



    
    /*This method executes a movement sequence if it's repeatedly called during the MOL. For now, it makes Leg 0 move its foot in a circle.
    * Returns true when complete.
    */
    boolean continueSequence(){
    	int sequenceTotalSteps = 30;//steps to divide the circle into 
    	double sequenceTotalTime = 1;//seconds to complete sequence
    	//foot positions in cm:
    	double centerX = -10;
    	double centerY = 0;
    	double centerZ = -9;
    	double radius = 5; 
    	sequenceTimer -= elapsed;
    	if(sequenceTimer <= 0){//we need to advance to next position
    		sequenceTimer = sequenceTotalTime/sequenceTotalSteps;//timer to next repositioning
    		//if(sequenceStep == 0)
    		//	sequenceTimer += 1;//add a 1 second delay to get to first position
    		//next position in the test circle in the constant Z plane:
    		double radians = (double)sequenceStep*(2*Math.PI)/sequenceTotalSteps;
    		Position pos = new Position();
    		pos.x = centerX+Math.sin(radians)*radius;
    		pos.y = centerY+Math.cos(radians)*radius;
    		pos.z = centerZ;
    		
    		moveLegTo(0, pos);
    		sequenceStep++;//this will be the next position
    	}
    	if(sequenceStep >= sequenceTotalSteps){
    		sequenceStep=0;
    		return true;//sequence is finished
    	}
    	else
    		return false;//not finished yet
    }
    /*constant X circle demo sequence:
      	double centerX = -32;
    	double centerY = 6;
    	double centerZ = -4;
    	double radius = 10; 
    	pos.x = centerX;
		pos.y = centerY+Math.cos(radians)*radius;
		pos.z = centerZ+Math.sin(radians)*radius;
    */
    
    
    
    //reposition the servo in response to user commands, and take user's report of the resulting joint angle.
    private void livecalibrate(){
    	liveServoGoalPositions[window.getLeg()][window.getServo()] = window.getSlider();//send the pulse width to the servo
    	servoGoalPositionChanged[window.getLeg()][window.getServo()] = true;
        //display commanded angles
        double dispangle = (window.getSlider()-servozero[window.getLeg()][window.getServo()])/servogain[window.getLeg()][window.getServo()];
        window.setText4(Math.round(dispangle) + " deg");

        if(window.saveCal == true){
            saveCalibration();
            window.saveCal = false;
            window.setText("Calibration data saved to file servocal.txt");
        }

        if(window.setPlusLim == true){
            servopluslimits[window.getLeg()][window.getServo()] = window.getSlider();
            window.setPlusLim = false;
        }
        if(window.setMinusLim == true){
            servominuslimits[window.getLeg()][window.getServo()] = window.getSlider();
            window.setMinusLim = false;
        }



        if(window.calAngle != -999){//means calibration angle has just been set
            if(calAngle1 == -999){//it hasn't been set
                calAngle1 = window.calAngle;
                calPulse1 = window.getSlider();
                window.calAngle = -999;//read the calibration angle and reset the input
                window.setText("first point saved, set new angle");
            }
            else{//angle 1 has been set, but angle 2 hasn't
                calAngle2 = window.calAngle;
                calPulse2 = window.getSlider();
                window.calAngle = -999;
                double gain = (double)(calPulse2 - calPulse1)/(calAngle2 - calAngle1);//find slope
                double zero = (double)calPulse1 - (calAngle1 * gain);//solve for b in pulse=m*angle + b
                servozero[window.getLeg()][window.getServo()] = zero;
                servogain[window.getLeg()][window.getServo()] = gain;
                calAngle1 = -999;
                calAngle2 = -999;//reset to be ready for a new calculation
                window.setText("Zero and gain calculated.");
            }
        }
    }

    //position a leg such that the foot is at the given position.
    //If the movement violates physics or joint limits, it does nothing and displays an error message.
    private void moveLegTo(int leg, Position command){
        JointAngles angles  = findJointAngles(leg, command, true);//elbow-up solution, for normal walking
        servoDigitalCommand theServoCommand = servoCommandFromAngles(leg, angles);
        if(theServoCommand.possible == false){//try elbow-down solution, for reaching up.
        	angles  = findJointAngles(leg, command, false);
            theServoCommand = servoCommandFromAngles(leg, angles);
            	if(theServoCommand.possible == false){//it still doesn't work.
            		window.setText("Impossible leg movement command received");
            		return;
            	}
        }
        for(int i=0; i<3; i++){
            liveServoGoalPositions[leg][i] = theServoCommand.goalPositions[i];
            servoGoalPositionChanged[leg][i] = true;
        }
        window.setText("Leg moved");
        //for debugging, print forward kinematic position for this command.
        Position theorPos = findFootPosition(leg,angles);
        print("x: " + roundTo2(theorPos.x) + "y: " + roundTo2(theorPos.y) + "z: " + roundTo2(theorPos.z) );
    }

    static double shoulderx[] = new double[6];//shoulder positions
    static double shouldery[] = new double[6];
    static double shoulderz[] = new double[6];//this actually needs to be the height of the joint at the end of segment 1
    static double seg1length[] = new double[6];//length of the part before the "up" joint
    static double seg2length[] = new double[6];
    static double seg3length[] = new double[6];

    
    //this method does the forward kinematics. Takes joint angles and returns the resulting foot position.
    //Used to figure out where the foot is now, or to numerically figure out the displacement from of a small movement.
    Position findFootPosition(int leg, double q1, double q2, double q3){//(my names in this function don't start at 0, because the D-H convention doesn't)
    	Position footPosition = new Position();
    	//The forward kinematic formulas were generated by the University of Illinois "robotica.m" extension to Mathematica.
    	footPosition.x = Math.cos(q1)* (seg1length[leg] + seg2length[leg]* Math.cos(q2) + seg3length[leg] * Math.cos(q2 + q3));
    	footPosition.y = (seg1length[leg] + seg2length[leg] * Math.cos(q2) + seg3length[leg] * Math.cos(q2 + q3)) * Math.sin(q1);
    	footPosition.z = seg2length[leg]* Math.sin(q2) + seg3length[leg]* Math.sin(q2 + q3);
    	//now we have position in the leg's coordinate frame; translate it to the robot coordinate frame.
    	footPosition.x += shoulderx[leg];
    	footPosition.y += shouldery[leg];
    	footPosition.z += shoulderz[leg];
    	return footPosition;
    }
    //wrapper
    Position findFootPosition(int leg, JointAngles angles){
    	return findFootPosition(leg, angles.q[0], angles.q[1], angles.q[2]);
    }
    
    //this method does the inverse kinematics. Takes a desired foot position and returns the required joint angles
    private JointAngles findJointAngles(int leg, Position command, boolean elbowUp){

        //first change command values into a reference frame where the shoulder of this leg is 0,0,0
        double x = command.x - shoulderx[leg];
        double y = command.y - shouldery[leg];
        double z = command.z - shoulderz[leg];
        double angle1 = Math.atan2(y,x);//this is arctan(y/x). 0 radians is the +x direction.
        //System.out.println(angle1*180/Math.PI);
        //note there's a discontinuity at the -x direction (the -pi to pi radians transition) which would mess up the angle-to-pulse-length conversion
        //below. But my legs never have to move through that direction, so I won't bother accounting for it yet.

        double R = Math.sqrt(x*x + y*y);//distance from shoulder to commanded point, neglecting z
        R = R - seg1length[leg];//really though, we care about the distance that has to be covered by the remaining 2 segments
        //now find the two vertical joint angles using Equations 1.7 and 1.8 of "Robot Modeling and Control" book
        //which derived this using the law of cosines. (Replace 'x' with R and 'y' with z in the equations.)
        double a1 = seg2length[leg];
        double a2 = seg3length[leg];
        double D = (R*R+z*z-a1*a1-a2*a2)/(2*a1*a2);
        double angle3;
        if(elbowUp){
        	angle3 = Math.atan2(-1*Math.sqrt(1-D*D), D);//This is the "elbow up" solution, which I'll use for the hexapod.
                                                            //the "elbow down" solution is given by removing the "-1"
        }else{
        	angle3 = Math.atan2(Math.sqrt(1-D*D), D);
        }
        double angle2 = Math.atan2(z, R) - Math.atan2( a2*Math.sin(angle3), a1 + a2*Math.cos(angle3));

        /*
        System.out.println("angle1: " + Math.round(angle1));
        System.out.println(angle1*180/Math.PI);
        System.out.println("angle2: " + Math.round(angle2));
        System.out.println(angle2*180/Math.PI);
        System.out.println("angle3: " + Math.round(angle3));
        System.out.println(angle3*180/Math.PI);
        */
        return new JointAngles(angle1, angle2, angle3);
    }
    
    private servoDigitalCommand servoCommandFromAngles(int leg, JointAngles angles){
    	servoDigitalCommand theServoCommand = new servoDigitalCommand();
    	theServoCommand.goalPositions[0] = (int)Math.round(angles.q[0]*180/Math.PI*servogain[leg][0] + servozero[leg][0]);
        theServoCommand.goalPositions[1] = (int)Math.round(angles.q[1]*180/Math.PI*servogain[leg][1] + servozero[leg][1]);
        theServoCommand.goalPositions[2] = (int)Math.round(angles.q[2]*180/Math.PI*servogain[leg][2] + servozero[leg][2]);
        theServoCommand.possible = true;
        for(int i = 0; i<3; i++){
            if(theServoCommand.goalPositions[i]>servopluslimits[leg][i] ||
                    theServoCommand.goalPositions[i]<servominuslimits[leg][i]){
                theServoCommand.possible = false;
                System.out.println("Requires movement beyond limit: Pulselength " + i +" was:");
                System.out.println(theServoCommand.goalPositions[i]);
            }
        }
        return theServoCommand;
    }
    
    private double angleFromServoPosition(int leg, int joint, int servoPosition){
    	return Math.toRadians((servoPosition - servozero[leg][joint])/servogain[leg][joint]);//inverse of the calculation in servoCommandFromAngles above.
    }

    private void saveCalibration() {
        try {
            PrintWriter outputStream = new PrintWriter(new FileWriter("servocal.txt"));
            outputStream.println("//Servo config file. Lines preceded by / will be ignored. File is wiped with each calibration.");
            for(int i=0;i<servopluslimits.length;i++){
                outputStream.println("//Leg " + i);
                for(int j=0; j<servopluslimits[0].length;j++){
                    outputStream.println("//Servo " + j + " + limit:");
                    outputStream.println(servopluslimits[i][j]);
                    outputStream.println("//- limit:");
                    outputStream.println(servominuslimits[i][j]);
                    outputStream.println("//zero:");
                    outputStream.println(servozero[i][j]);
                    outputStream.println("//gain:");
                    outputStream.println(servogain[i][j]);
                }
            }
            outputStream.close();
        }
        catch (IOException e) {
            window.setText("File problem; could not save.");
        }
    }

    private void initRobotData(){
    	//temporarily hacking this in. It will be read from a file eventually.
    	
    	//Positions of joint 1 for each leg. (This is used as the origin of the leg's coordinate system)
        shoulderx[0] = 0;
        shouldery[0] = 0;
        shoulderz[0] = 0;
        
        //Lengths of each leg segment, from joint to joint (or joint to foot)
        
        //AX-12 leg values (the leg built using the AX-12's included attachment hardware)
        seg1length[0] = 7;//from "shoulder" rotational joint to up-down joint
        seg2length[0] = 5.3;//from first up-down joint to second 
        seg3length[0] = 6.8;//from second joint to foot
        
        /*//old RC servo leg values:
        seg1length[0] = 8;//from rotational joint to up-down joint
        seg2length[0] = 13.8;
        seg3length[0] = 22.6;
         */
    }

    private boolean loadCalibration() {
        try {
            BufferedReader inputStream = new BufferedReader(new FileReader("servocal.txt"));
            String line = null;

            for(int i=0;i<servopluslimits.length;i++){
                for(int j=0; j<servopluslimits[0].length;j++){
                    while((line = inputStream.readLine()).charAt(0)=='/')//ignore the comment lines
                    {}
                    servopluslimits[i][j] = Integer.parseInt(line);
                    while((line = inputStream.readLine()).charAt(0)=='/')
                    {}
                    servominuslimits[i][j] = Integer.parseInt(line);
                    while((line = inputStream.readLine()).charAt(0)=='/')
                    {}
                    servozero[i][j] = Double.parseDouble(line);
                    while((line = inputStream.readLine()).charAt(0)=='/')
                    {}
                    servogain[i][j] = Double.parseDouble(line);
                    }
                }
            inputStream.close();
            return true;
            }
        catch (IOException e) {
            window.setText("Could not load previous servo calibration data.");
            return false;
        }
    }
    
  //transmitDigitalCommand() is a custom method that transmits the angles to AX-12 servos, using the USB2Dynamixel API.
    //@SuppressWarnings("unused")
	private void transmitDigitalCommand(){
    	//----forward servo pulse times to the microcontroller, to be applied to the servos.
    	//System.out.println("Writing \""+messageString+"\" to "
    	//		+serialPort.getName());
        for(int i=0; i<liveServoGoalPositions.length;i++){//'i' is the leg, 0-5
            for(int j=0; j<liveServoGoalPositions[0].length; j++){//'j' is the servo within the leg, 0-2
            	if(servoGoalPositionChanged[i][j]){//we have a new position to command
	            	int id = i*3+j;
	                dynamixel_jni.dxl_write_word(id, GOAL_POSITION, liveServoGoalPositions[i][j]);
            	}
            }
        }
    }
	
	void readServoData(){
		int leg = 0;//replace with an outer loop when I build all 6 legs
		for(int joint = 0; joint<3; joint++){
			int servoLoad = 0;
			int id = leg*3+joint;
			servoLoad = dynamixel_jni.dxl_read_word( id , PRESENT_LOAD );
			if( dynamixel_jni.dxl_get_result( ) == dynamixel_jni.COMM_RXSUCCESS )
			{
				final int AX12FULLTORQUE = 0x03FF;
				int signedLoad=0;
				if((servoLoad & (1<<10)) == 0 ){//bit 10 is 0 for CCW, 1 for CW.
					signedLoad = -1*(servoLoad & AX12FULLTORQUE);//bits 9 thru 0 are the value.
				}else{
					signedLoad = (servoLoad & AX12FULLTORQUE);
				}
				double percentLoad = (double)signedLoad/AX12FULLTORQUE;//fraction of full torque currently being applied
				//now convert fraction of torque into Newton-meters, while considering the direction.
				if(servogain[leg][joint]>0)//if calibrated gain is positive, servo "+" direction matches my angle convention.
					measuredJointTorques[leg][joint] = percentLoad*FULLTORQUENM;
				else
					measuredJointTorques[leg][joint] = -percentLoad*FULLTORQUENM;//servo "+" direction is my convention's "-", so change the sign.
				/*percentLoad = (Math.round(percentLoad*100))/100f;//round to 2 decimals.
				window.loadLabel.setText("Load is: " + percentLoad);*/
			}
			int servoPosition = 0;
			servoPosition = dynamixel_jni.dxl_read_word( id , PRESENT_POSITION );
			if( dynamixel_jni.dxl_get_result( ) == dynamixel_jni.COMM_RXSUCCESS ){
				measuredJointAngles[leg][joint] = angleFromServoPosition(leg, joint, servoPosition);
			}
		}
	}
	
	//This method uses the previously-read joint torque and position values, stored in measuredJointTorques and measuredJointAngles arrays.
	//It calculates the force which is being applied to the end-effector (a.k.a. the foot) given those torques.
	@SuppressWarnings("deprecation")
	void findForce(boolean verbose){
		int leg = 0;
		if(verbose)
			print("leg " + leg + " q0: " + Math.toDegrees(measuredJointAngles[leg][0]) + 
				" q1: " +Math.toDegrees(measuredJointAngles[leg][1]) + 
				" q2: " + Math.toDegrees(measuredJointAngles[leg][2]) );
		
		final double INFINITESIMAL = 0.001;//an arbitrary small change in radians that we'll use to numerically find the derivative
		JointAngles currentAngles = new JointAngles(measuredJointAngles[leg][0],measuredJointAngles[leg][1],measuredJointAngles[leg][2]);
		Position currentPos = findFootPosition(leg, currentAngles);
		
		//jacobianT is a matrix of conversion factors, equivalent to transpose of jacobian
		RealMatrix jacobianT = new RealMatrixImpl(3, 3);//each row will hold factors that convert delta-q to delta-x, delta-y, delta-z 
		for(int joint=0; joint<3; joint++){
			JointAngles displacedAngles = new JointAngles(currentAngles);//copy current angles
			displacedAngles.q[joint] += INFINITESIMAL;// and perturb only the current joint; leave the rest.
			Position displacedPos = findFootPosition(leg, displacedAngles);//find foot position after that perturbation
			Position deltaP = displacedPos.minus(currentPos);//find change caused by perturbation
			Position factors = deltaP.times(1/INFINITESIMAL * 1.0/100.0 );//find delta-P/delta-Q, a.k.a. the a numerical derivative of
								//the forward kinematics with respect to this joint. This is the same as a column of the Jacobian matrix but only
								//for speed, not angular velocity. Also divide by 100 to convert cm to meters.
			jacobianT.setRow(joint, factors.getArray() );
			if(verbose)
				print("joint " + joint + " gx: " + factors.x +" gy: " + factors.y +" gz: " + factors.z);
		}
		
		
		//now that our Jacobian transpose is set for the current location, take the inverse of it.
		if(new LUDecompositionImpl(jacobianT).getSolver().isNonSingular()){//otherwise, if matrix is singular, we can't take the inverse, so do nothing.
													//The matrix will be singular at the edge of the range of motion and where the foot is directly
													//above or below the shoulder.
			jacobianIsNonSingular = true;
			RealMatrix JTinv = new LUDecompositionImpl(jacobianT).getSolver().getInverse();
			RealMatrix torques = new RealMatrixImpl(measuredJointTorques[leg]);
			RealMatrix forces = torques.preMultiply(JTinv);
			measuredForces[leg] = forces.getColumn(0);//this is our final output
			
			/** Force error determination:
			 * perturb torque by an amount equal to its accuracy, and remember how much calculated force changes.
			 * This is done for each joint, and max error in force direction is saved in "measuredForceError".
			 */
			measuredForceError[leg][0] = 0;//note that the second index here is the force axis (x,y,z) not the joint number.
			measuredForceError[leg][1] = 0;
			measuredForceError[leg][2] = 0;
			for(int joint=0; joint<3; joint++){//for each joint, perturb torque and find resulting change in force. Add it to the force error in each axis.
				torques.addToEntry(joint, 0, TORQUEACCURACY);
				RealMatrix perturbedForces = torques.preMultiply(JTinv);
				RealMatrix deltaForce =  perturbedForces.add(forces.scalarMultiply(-1));
				for(int axis=0; axis<3; axis++){
					measuredForceError[leg][axis] += Math.abs(deltaForce.getEntry(axis, 0));
				}
				//restore torque (minor floating point error is acceptable here)
				torques.addToEntry(joint, 0, -TORQUEACCURACY);
			}
			//********* end of force error determination
			
			
			//If error means that force could actually be zero, mark that force direction as unreliable, for convenience.
			measuredForceIsCertainlyNonzero[leg][0]=true;
			measuredForceIsCertainlyNonzero[leg][1]=true;
			measuredForceIsCertainlyNonzero[leg][2]=true;
			for(int axis=0; axis<3; axis++){
				if(Math.abs(measuredForces[leg][axis]) - measuredForceError[leg][axis]  < 0)
					measuredForceIsCertainlyNonzero[leg][axis]=false;
			}
			
			//now run a low-pass filter
			final double alpha = .2;//smaller alpha increases effective filtering time
			for(int axis=0; axis<3; axis++){
				//if(measuredForcesAreReliable[leg][joint]){
				measuredForcesFiltered[leg][axis] = measuredForcesFiltered[leg][axis]*(1-alpha) + alpha*measuredForces[leg][axis]; 
				double forceBound = FULLTORQUENM*10000;
				measuredForcesFiltered[leg][axis] = Math.min(Math.max(-forceBound, measuredForcesFiltered[leg][axis]), forceBound);
				//}
			}
		}else{//matrix was singular
			jacobianIsNonSingular = false;
			measuredForceIsCertainlyNonzero[leg][0]=false;
			measuredForceIsCertainlyNonzero[leg][1]=false;
			measuredForceIsCertainlyNonzero[leg][2]=false;
		}
		//end of functional code
				
		//begin user interface output:
		
		
		if(verbose)
			print("Forces - x: " + measuredForcesFiltered[leg][0] + " y: " + measuredForcesFiltered[leg][1] + " z: " + measuredForcesFiltered[leg][2]);
		
		DecimalFormat twoDForm = new DecimalFormat("#.##");
		String message = "Reported Loads: ";
		for(int joint=0; joint<3; joint++)
			message += twoDForm.format(measuredJointTorques[leg][joint]) + ", ";
		String xmessage = "  x: " + twoDForm.format(measuredForcesFiltered[leg][0]);
		String ymessage = "  y: " + twoDForm.format(measuredForcesFiltered[leg][1]);
		String zmessage = "  z: " + twoDForm.format(measuredForcesFiltered[leg][2]);
		if(measuredForceIsCertainlyNonzero[leg][0])
			xmessage += "#";
		if(measuredForceIsCertainlyNonzero[leg][1])
			ymessage += "#";
		if(measuredForceIsCertainlyNonzero[leg][2])
			zmessage += "#";
		String message2 = "Calculated Force:" + xmessage + ymessage + zmessage;
		window.loadLabel.setText(message);
		window.lblCalculatedForces.setText(message2);
	}
	
	
	//turn off the servos, so they'll let themselves be repositioned by hand. Do this on exit or in a safety situation.
	private void torqueoff(){
		int leg=0;
		//for(int leg=0; leg<liveServoGoalPositions.length;leg++){//'i' is the leg, 0-5
		
            for(int joint=0; joint<liveServoGoalPositions[0].length; joint++){//'j' is the servo within the leg, 0-2
            	int id = leg*3+joint;
                dynamixel_jni.dxl_write_byte(id, TORQUE_ENABLE, 0);
            }
		//}
	}
	
	//set my desired default values for the RAM constants in all servos
	private void configServosForTests(){
        for(int i=0; i<liveServoGoalPositions.length;i++){//'i' is the leg, 0-5
            for(int j=0; j<liveServoGoalPositions[0].length; j++){//'j' is the servo within the leg, 0-2
            	int id = i*3+j;
                dynamixel_jni.dxl_write_byte(id, ALARM_LED, 0xff);
                dynamixel_jni.dxl_write_byte(id, ALARM_SHUTDOWN, 0xff);
                dynamixel_jni.dxl_write_byte(id, TORQUE_ENABLE, 0x01);
                dynamixel_jni.dxl_write_word(id, TORQUE_LIMIT, 512);//for calibration, set torque limit to half, whatever that means.
            }
        }
        System.out.println("Servos are set to low torque for testing.");
    }
    private void print(String s){
    	window.print(s);
    }
    String roundTo2(double number){
    	DecimalFormat twoDForm = new DecimalFormat("#.##");
    	return twoDForm.format(number);
    }
/*
    private void openSerial(){
        boolean portFound = false;
	String  defaultPort = "COM1";

	portList = CommPortIdentifier.getPortIdentifiers();

	while (portList.hasMoreElements()) {
	    portId = (CommPortIdentifier) portList.nextElement();

	    if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {

		if (portId.getName().equals(defaultPort)) {
		    System.out.println("Found port " + defaultPort);

		    portFound = true;

		    try {
			serialPort =
			    (SerialPort) portId.open("SimpleWrite", 2000);
		    } catch (PortInUseException e) {
			System.out.println("Port in use.");

			continue;
		    }

		    try {
			outputStream = serialPort.getOutputStream();
		    } catch (IOException e) {}

		    try {
			serialPort.setSerialPortParams(115200,//baud rate. 115200 is the fastest the PIC supports, 9600 is the slowest.
						       SerialPort.DATABITS_8,
						       SerialPort.STOPBITS_1,
						       SerialPort.PARITY_NONE);
		    } catch (UnsupportedCommOperationException e) {}

		    try {
		    	serialPort.notifyOnOutputEmpty(true);
		    } catch (Exception e) {
			System.out.println("Error setting event notification");
			System.out.println(e.toString());
			System.exit(-1);
		    }
		}
	    }
	}

	if (!portFound) {
	    System.out.println("port " + defaultPort + " not found.");
	}
    }

    public void closeSerial(){
        serialPort.close();
    }
    
    
    
    //transmitPWMToMicrocontroller() is a custom method that talks to my microcontroller code to give it the desired servo pulse times.
    //for efficiency I express them as 0-255, which the microcontroller will scale to 0.5-2.5 ms repeating every 20ms
    @SuppressWarnings("unused")
	private void transmitPWMCommand(){
    	//----forward servo pulse times to the microcontroller, to be applied to the servos.
    	//System.out.println("Writing \""+messageString+"\" to "
    	//		+serialPort.getName());
        byte[] transmission = new byte[liveservotimes.length * liveservotimes[0].length];
        for(int i=0; i<liveservotimes.length;i++){
            for(int j=0; j<liveservotimes[0].length; j++){
                transmission[i*liveservotimes[0].length + j] = (byte)liveservotimes[i][j];//I'll need to redo this if I want more than 0-255
            }
        }
		try {
			outputStream.write("go".getBytes());//This signifies the start of the message string.
	                                                    //(The microcontroller may get turned on in the middle, so it needs to get synced up.)
	                outputStream.write(transmission);
		} catch (IOException e){
		}
    }
*/

}


class Position{//This stores a desired robot foot position. (My code assumes the position is expressed in the robot's coordinate frame)
    public double x;
    public double y;
    public double z;

    Position(){

    }
    Position(double inputx, double inputy, double inputz){
        x = inputx;
        y = inputy;
        z = inputz;
    }
    public Position minus(Position in){
    	return new Position(this.x-in.x,this.y-in.y,this.z-in.z );
    }
    public Position plus(Position in){
    	return new Position(this.x+in.x,this.y+in.y,this.z+in.z );
    }
    public Position times(double scalar){
    	return new Position(this.x*scalar,this.y*scalar,this.z*scalar);
    }
    public double[] getArray(){
    	double[] coordList = new double[3];
    	coordList[0] = x;
    	coordList[1] = y;
    	coordList[2] = z;
    	return coordList;
    }
}
class JointAngles{//This stores joint angles of one leg
	public double[] q;

    JointAngles(){
    	q = new double[3];
    }
    JointAngles(double a1in, double a2in, double a3in){
    	q = new double[3];
        q[0] = a1in;
        q[1] = a2in;
        q[2] = a3in;
    }
    JointAngles(JointAngles orig){
    	q = new double[3];
        q[0] = orig.q[0];
        q[1] = orig.q[1];
        q[2] = orig.q[2];
    }
}
//This contains the pulse widths for one robot leg worth of standard hobby servos, as well as
//a boolean stating whether the move can be done, given leg lengths and
//joint angle limits. (Deprecated when I switched to AX-12 servos)
class servoPWMCommand{
	 public int servopulse[];
	 public boolean possible;
	 servoPWMCommand(){
		 servopulse = new int[3];
	 }
}

//This contains the pulse widths for one robot leg worth of digital servos, as well as
//a boolean stating whether the move can be done, given leg lengths and
//joint angle limits.
class servoDigitalCommand{
	 public int goalPositions[];
	 public boolean possible;
	 servoDigitalCommand(){
		 goalPositions = new int[3];
	 }
}